module SynOps where

import List(union,(\\))
import HsSyn

import IOExts(trace)
deb s e = length sx `seq` trace sx e
            where sx = s ++ ": " ++ show e

class MapExp s where
   mapExp :: ([HsName] -> HsExp -> HsExp) -> [HsName] -> s -> s

instance MapExp s => MapExp [s] where
   mapExp f n = map (mapExp f n)

instance MapExp HsModule where
    mapExp f n (HsModule m e i ds) = HsModule m e i (mapExpDecls f n ds)

--instance MapExp HsDecl where
mapExpDecl f n (HsClassDecl s t ds) = HsClassDecl s t (mapExpDecls f n ds)
mapExpDecl f n (HsInstDecl s t ds)  = HsInstDecl s t (mapExpDecls f n ds)
mapExpDecl f n (HsFunBind s m)      = HsFunBind s (mapExp f n m)
mapExpDecl f n (HsPatBind s p r ds) = HsPatBind s p (mapExp f n' r) 
                                                    (mapExpDecls f n ds)
                                  where n' = n `union` defined ds
mapExpDecl f n d                    = d

mapExpDecls f n ds = map (mapExpDecl f (n `union` defined ds)) ds

instance MapExp HsMatch where
  mapExp f n (HsMatch s m ps r ds) = HsMatch s m ps (mapExp f n' r) 
                                                    (mapExpDecls f n ds)
                                    where n' = n `union` fv ps
                                                 `union` defined ds

instance MapExp HsRhs where
  mapExp f n (HsUnGuardedRhs e) = HsUnGuardedRhs (mapExp f n e)
  mapExp f n (HsGuardedRhss gs) = HsGuardedRhss (mapExp f n gs)

instance MapExp HsGuardedRhs where
  mapExp f n (HsGuardedRhs s e1 e2) = HsGuardedRhs s (mapExp f n e1)
                                                     (mapExp f n e2)

instance MapExp HsExp where
  mapExp f n e = f n e

instance MapExp ArrowCmd where
  mapExp f n (a :-< b)         = mapExp f n a :-< mapExp f n b
  mapExp f n (ArForm e cs)     = ArForm (mapExp f n e) (mapExp f n cs)
  mapExp f n (ArInfixOp c e d) = ArInfixOp (mapExp f n c) (mapExp f n e) 
                                           (mapExp f n d)
  mapExp f n (ArKappa p c)     = ArKappa p (mapExp f (n `union` fv p) c)
  mapExp f n (ArExp e)         = ArExp (mapExp f n e)
  mapExp f n (ArLet ds c)      = ArLet (mapExpDecls f n ds) (mapExp f n' c)
                                 where n' = n `union` defined ds
  mapExp f n (ArDo (ss,c))     = ArDo (mapExp f n ss,mapExp f n c)

instance MapExp ArrowStmt where
  mapExp f n (ArBind p c) = ArBind p (mapExp f (n `union` fv p) c)
  mapExp f n (ArLetStmt ds) = ArLetStmt (mapExpDecls f n ds)
  mapExp f n (ArRec ss) = ArRec (mapExp f (n `union` defined ss) ss)

instance MapExp HsStmt where
  mapExp f n (HsGenerator p e) = HsGenerator p (mapExp f (n `union` fv p) e)
  mapExp f n (HsQualifier e) = HsQualifier (mapExp f n e)
  mapExp f n (HsLetStmt ds) = HsLetStmt (mapExpDecls f n ds)

instance MapExp HsFieldUpdate where
  mapExp f n (HsFieldUpdate m e) = HsFieldUpdate m (mapExp f n e)

instance MapExp HsAlt where
  mapExp f n (HsAlt s p gs ds) = HsAlt s p (mapExp f n' gs) 
                                           (mapExpDecls f n ds)
                                 where n' = n `union` fv p
                                              `union` defined ds

instance MapExp HsGuardedAlts where
  mapExp f n (HsUnGuardedAlt e) = HsUnGuardedAlt (mapExp f n e)
  mapExp f n (HsGuardedAlts as) = HsGuardedAlts (mapExp f n as)

instance MapExp HsGuardedAlt where
  mapExp f n (HsGuardedAlt s a b) = HsGuardedAlt s (mapExp f n a) 
                                                   (mapExp f n b)
mapExpPart f n = me where
   me :: MapExp s => s -> s
   me = mapExp fp n
   fp n e = 
       case f n e of
        Just e' -> e'
        Nothing -> case e of
           HsInfixApp a b c -> HsInfixApp (me a) (me b) (me c)
           HsApp a b -> HsApp (me a) (me b)
           HsNegApp e -> HsNegApp (me e)
           HsLambda ps e -> HsLambda ps (mapExp fp (n `union` fv ps) e)
           HsLet ds e -> HsLet (mapExpDecls fp n ds) (mapExp fp n' e)
                          where n' = n `union` defined ds
           HsIf a b c -> HsIf (me a) (me b) (me c)
           HsCase a as -> HsCase (me a) (me as)
           HsDo ss -> HsDo (me ss)
           HsTuple ss -> HsTuple (me ss)
           HsList es -> HsList (me es)
           HsParen e -> HsParen (me e)
           HsLeftSection a b -> HsLeftSection (me a) (me b)
           HsRightSection a b -> HsRightSection (me a) (me b)
           HsRecConstr n fs -> HsRecConstr n (me fs)
           HsRecUpdate e fs -> HsRecUpdate (me e) (me fs)
           HsEnumFrom e -> HsEnumFrom (me e)
           HsEnumFromTo a b -> HsEnumFromTo (me a) (me b)
           HsEnumFromThen a b -> HsEnumFromThen (me a) (me b)
           HsEnumFromThenTo a b c -> HsEnumFromThenTo (me a) (me b) (me c)
           HsListComp a ss -> HsListComp (me a) (me ss)
           HsExpTypeSig s e t -> HsExpTypeSig s (me e) t
           HsArrowProc p c -> HsArrowProc p (mapExp fp (n `union` fv p) c)
           a :* b -> me a :* me b
           e -> e -- deb "mapExpPart" e

class Fv p where
    fv :: p -> [HsName]

instance Fv HsExp where
  fv (HsVar n) = [n]
  fv (HsCon n) = [n]
  fv (HsInfixApp a b c) = fv a `union` fv b `union` fv c
  fv (HsApp a b) = fv a `union` fv b
  fv (HsNegApp a) = fv a
  fv (HsLambda ps a) = fv a \\ fv ps
  fv (HsLet ds a) = (fv a `union` fvHsDecls ds) \\ defined ds
  fv (HsIf a b c) =  fv a `union` fv b `union` fv c
  fv (HsCase a as) = fv a `union` fv as
  fv (HsDo  ss) = fvHsStmt ss
  fv (HsTuple as) = fv as
  fv (HsList as) = fv as
  fv (HsParen a) = fv a
  fv (HsLeftSection a b) = fv a `union` fv b
  fv (HsRightSection a b) = fv a `union` fv b
  fv (HsRecConstr n us) = [n] `union` fv us
  fv (HsRecUpdate a us) = fv a `union` fv us
  fv (HsEnumFrom a) = fv a
  fv (HsEnumFromTo a b) = fv a `union` fv b
  fv (HsEnumFromThen a b) = fv a `union` fv b
  fv (HsEnumFromThenTo a b c) = fv a `union` fv b `union` fv c
  fv (HsListComp a ss) = fv a `union` fvHsStmt ss
  fv (HsExpTypeSig s a t) = fv a
  fv (HsArrowProc p c) = fv c \\ fv p
  fv (a :* b) = fv a `union` fv b
  fv _ = []

class Defined p where
  defined :: p -> [HsName]

instance Defined p => Defined [p] where
  defined = foldr union [] . map defined

instance Fv a => Fv [a] where
  fv = foldr union [] . map fv

instance Defined HsMatch where
  defined (HsMatch s n ps rs ds) = [n]

instance Defined HsDecl where
  defined (HsClassDecl s q ds) = defined ds
  defined (HsInstDecl s q ds) = defined ds
  defined (HsFunBind s m) = defined m
  defined (HsPatBind s p r ds) = fv p
  defined _ = []

instance Defined ArrowStmt where
  defined (ArBind p c) = fv p
  defined (ArLetStmt ds) = defined ds
  defined (ArRec rs) = defined rs

instance Fv HsPat where
  fv (HsPVar n) = [n]
  fv (HsPLit l) = []
  fv (HsPNeg p) = fv p
  fv (HsPInfixApp p n q) = fv p `union` [n] `union` fv q
  fv (HsPApp n ps) = [n] `union` fv ps
  fv (HsPTuple ps) = fv ps
  fv (HsPList ps) = fv ps
  fv (HsPParen p) = fv p
  fv (HsPRec n ps) = [n] `union` fv ps
  fv (HsPAsPat n p) = [n] `union` fv p
  fv HsPWildCard = []
  fv (HsPIrrPat p) = fv p

fvHsDecls ds = foldr union [] (map fvHsDecl ds) \\ defined ds
 where
  fvHsDecl (HsClassDecl s q ds) = fvHsDecls ds
  fvHsDecl (HsInstDecl s q ds) = fvHsDecls ds
  fvHsDecl (HsFunBind s m) = fv m
  fvHsDecl (HsPatBind s p r ds) = (fv r `union` fvHsDecls ds) \\ fv p
  fvHsDecl _ = []

instance Fv HsMatch where
  fv (HsMatch s n ps rs ds) = ((fv rs \\ defined ds) `union` fvHsDecls ds) \\ 
                              ([n] `union` fv ps)

instance Fv HsRhs where
  fv (HsUnGuardedRhs e) = fv e
  fv (HsGuardedRhss gs) = fv gs

instance Fv HsGuardedRhs where
  fv (HsGuardedRhs s a b) = fv a `union` fv b

instance Fv HsPatField where
  fv (HsPFieldPat n p) = fv p

instance Fv HsAlt where
  fv (HsAlt s p as ds) = (fv as `union` fvHsDecls ds) \\ fv p

instance Fv HsGuardedAlts where
  fv (HsUnGuardedAlt e) = fv e
  fv (HsGuardedAlts as) = fv as

instance Fv HsGuardedAlt where
  fv (HsGuardedAlt s a b) = fv a `union` fv b

fvHsStmt [] = []
fvHsStmt (HsGenerator p a:ss) = fv a `union` (fvHsStmt ss \\ fv p)
fvHsStmt (HsQualifier a:ss) = fv a `union` fvHsStmt ss
fvHsStmt (HsLetStmt ds:ss) = (fvHsDecls ds `union` fvHsStmt ss) \\ defined ds

instance Fv HsFieldUpdate where
  fv (HsFieldUpdate n e) = [n] `union` fv e

instance Fv ArrowCmd where
  fv (a :-< b) = fv a `union` fv b
  fv (ArForm a ss) = fv a `union` fv ss
  fv (ArInfixOp a b c) = fv a `union` fv b `union` fv c
  fv (ArKappa p c) = fv c \\ fv p
  fv (ArExp a) = fv a
  fv (ArLet ds c) = fvHsDecls ds `union` (fv c \\ defined ds)
  fv (ArDo ss) = fvArrowStmts ss

fvArrowStmts ([],c) = fv c
fvArrowStmts (s:ss,c) = fvArrowStmt s `union` (fvArrowStmts (ss,c) 
                                               \\ defined s)
  where
   fvArrowStmt (ArBind p c) = fv c
   fvArrowStmt (ArLetStmt ds) = fvHsDecls ds
   fvArrowStmt (ArRec rs) = foldr union [] (map fvArrowStmt rs) \\ defined rs

instance Fv ArrowPat where
  fv (p :*: q) = fv p `union` fv q
  fv (ArPParen p) = fv p
  fv (ArHsPat p) = fv p

class P2E a where
   p2e :: a -> HsExp

instance P2E HsPat where
  p2e (HsPVar n) = HsVar n
  p2e (HsPLit l) = HsLit l
  p2e (HsPNeg p) = HsNegApp (p2e p)
  p2e (HsPInfixApp p n q) = HsInfixApp (p2e p) (HsVar n) (p2e q)
  p2e (HsPApp n ps) = foldl HsApp (HsVar n) (map p2e ps)
  p2e (HsPTuple ps) = HsTuple (map p2e ps)
  p2e (HsPList ps) = HsList (map p2e ps)
  p2e (HsPParen p) = HsParen (p2e p)
  p2e _ = undefined

instance P2E ArrowPat where
  p2e (p :*: q) = p2e p :* p2e q
  p2e (ArPParen p) = HsParen (p2e p)
  p2e (ArHsPat p) = p2e p

class ZapVar p where
  zapVar :: [HsName] -> p -> p

instance ZapVar ArrowPat where
  zapVar ns (p1 :*: p2)  = zapVar ns p1 :*: zapVar ns p2
  zapVar ns (ArPParen p) = ArPParen (zapVar ns p)
  zapVar ns (ArHsPat p)  = ArHsPat (zapVar ns p)

instance ZapVar p => ZapVar [p] where
  zapVar ns = map (zapVar ns)

instance ZapVar HsPat where
  zapVar ns (HsPVar n)      | n `elem` ns = HsPWildCard
  zapVar ns (HsPNeg p)                    = HsPNeg (zapVar ns p)
  zapVar ns (HsPInfixApp p1 n p2)         = HsPInfixApp (zapVar ns p1) n 
                                                       (zapVar ns p2)
  zapVar ns (HsPApp n ps)                 = HsPApp n (zapVar ns ps)
  zapVar ns (HsPTuple ps)                 = HsPTuple (zapVar ns ps)
  zapVar ns (HsPList ps)                  = HsPList (zapVar ns ps)
  zapVar ns (HsPParen p)                  = HsPParen (zapVar ns p)
  zapVar ns (HsPRec n ps)                 = HsPRec n (zapVar ns ps)
  zapVar ns (HsPAsPat n ps) | n `elem` ns = zapVar ns ps
                            | otherwise   = HsPAsPat n (zapVar ns ps)
  zapVar ns (HsPIrrPat p)                 = HsPIrrPat (zapVar ns p)
  zapVar ns p                             = p

instance ZapVar HsPatField where
  zapVar ns (HsPFieldPat n p) = HsPFieldPat n (zapVar ns p)
